@@ -1,6 +1,6 @@
-_MAJOR = 3
-_MINOR = 2
-_PATCH = "0.post1"
+_MAJOR=3
+_MINOR=7
+_PATCH="2"
 
 VERSION = f"{_MAJOR}.{_MINOR}.{_PATCH}"
 USER_VERSION = f"""{_MAJOR}.{_MINOR}.{_PATCH[0:1]} {
@@ -9,4 +9,4 @@ USER_VERSION = f"""{_MAJOR}.{_MINOR}.{_PATCH[0:1]} {
     f'beta {_PATCH[_PATCH.rindex("b") + 1:]}' if 'b' in _PATCH else ''
 }{
     f'release candidate {_PATCH[_PATCH.rindex("rc") + 2:]}' if 'rc' in _PATCH else ''
-}"""
+}"""
\ No newline at end of file
