@@ -11,6 +11,7 @@ from pathlib import Path
 from platform import system
 from tempfile import TemporaryDirectory
 from typing import Optional, Union, List, Tuple
+from zlib import crc32
 
 import oead
 import xxhash  # pylint: disable=wrong-import-order
@@ -20,7 +21,7 @@ from bcml import util, install
 from bcml.util import BYML_EXTS, SARC_EXTS, TempSettingsContext
 from bcml.mergers.pack import SPECIAL
 
-EXCLUDE_EXTS = {".yml", ".yaml", ".bak", ".txt", ".json", ".old"}
+EXCLUDE_EXTS = {".yml", ".yaml", ".bak", ".txt", ".json", ".old", ".bnp"}
 
 
 def _yml_to_byml(file: Path):
@@ -40,6 +41,30 @@ def _yml_to_aamp(file: Path):
     file.unlink()
 
 
+def _package_code(tmp_dir: Path, meta: dict):
+    patch_file: Optional[Path] = None
+    if (tmp_dir / "patches.txt").exists():
+        patch_file = tmp_dir / "patches.txt"
+    elif list(tmp_dir.glob("patch_*.asm")):
+        patch_file = list(tmp_dir.glob("patch_*.asm"))[0]
+    if not patch_file and not (tmp_dir / "patches").exists():
+        return
+    (tmp_dir / "patches").mkdir(exist_ok=True)
+    shutil.move(patch_file, tmp_dir / "patches" / patch_file.name)
+    if not (tmp_dir / "patches" / "rules.txt").exists():
+        rules = {
+            "titleIds": "00050000101C9300,00050000101C9400,00050000101C9500",
+            "name": meta["name"] + " - Executable Patch",
+            "path": f'"The Legend of Zelda: Breath of the Wild/Mods/{meta["name"]} Patch (BCML)"',
+            "description": f"Contains executable patch(es) for {meta['name']}",
+            "version": 4,
+        }
+        nl = "\n"
+        (tmp_dir / "patches" / "rules.txt").write_text(
+            f"[Definition]{nl}{nl.join(f'{k} = {v}' for k, v in rules.items())}"
+        )
+
+
 def _pack_sarcs(tmp_dir: Path, hashes: dict, pool: multiprocessing.pool.Pool):
     sarc_folders = {
         d
@@ -98,9 +123,7 @@ def _pack_sarc(folder: Path, tmp_dir: Path, hashes: dict):
                 old_hash = xxhash.xxh64_intdigest(
                     util.unyaz_if_needed(old_sarc.get_file(file_name).data)
                 )
-            if file_name not in old_files or (
-                xhash != old_hash and file.suffix not in util.AAMP_EXTS
-            ):
+            if file_name not in old_files or (xhash != old_hash):
                 packed.files[file_name] = file_data
     finally:
         shutil.rmtree(folder)
@@ -114,11 +137,14 @@ def _pack_sarc(folder: Path, tmp_dir: Path, hashes: dict):
         )
 
 
+CLEAN_EXTS = util.SARC_EXTS - {".beventpack", ".sbeventpack"}
+
+
 def _clean_sarcs(tmp_dir: Path, hashes: dict, pool: multiprocessing.pool.Pool):
     sarc_files = {
         file
         for file in tmp_dir.rglob("**/*")
-        if file.suffix in util.SARC_EXTS
+        if file.suffix in CLEAN_EXTS
         and "options" not in file.relative_to(tmp_dir).parts
     }
     if sarc_files:
@@ -128,12 +154,12 @@ def _clean_sarcs(tmp_dir: Path, hashes: dict, pool: multiprocessing.pool.Pool):
     sarc_files = {
         file
         for file in tmp_dir.rglob("**/*")
-        if file.suffix in util.SARC_EXTS
+        if file.suffix in CLEAN_EXTS
         and "options" not in file.relative_to(tmp_dir).parts
     }
     if sarc_files:
         print("Updating pack log...")
-        final_packs = [file for file in sarc_files if file.suffix in util.SARC_EXTS]
+        final_packs = [file for file in sarc_files if file.suffix in CLEAN_EXTS]
         if final_packs:
             (tmp_dir / "logs").mkdir(parents=True, exist_ok=True)
             (tmp_dir / "logs" / "packs.json").write_text(
@@ -178,7 +204,7 @@ def _clean_sarc(old_sarc: oead.Sarc, base_sarc: oead.Sarc) -> Optional[oead.Sarc
             file_data != old_data and ext not in util.AAMP_EXTS
         ):
             if (
-                ext in util.SARC_EXTS
+                ext in CLEAN_EXTS
                 and nest_file in old_files
                 and nest_file not in SPECIAL
             ):
@@ -306,11 +332,19 @@ def create_bnp_mod(mod: Path, output: Path, meta: dict, options: Optional[dict]
     if not (
         (tmp_dir / util.get_content_path()).exists()
         or (tmp_dir / util.get_dlc_path()).exists()
+        or (tmp_dir / "patches").exists()
     ):
         if (tmp_dir.parent / util.get_content_path()).exists():
             tmp_dir = tmp_dir.parent
         elif util.get_settings("wiiu") and (tmp_dir / "Content").exists():
             (tmp_dir / "Content").rename(tmp_dir / "content")
+        elif not util.get_settings("wiiu") and (
+            (tmp_dir / "base").exists() or (tmp_dir / "dlc").exists()
+        ):
+            if (tmp_dir / "base").exists():
+                (tmp_dir / "base").rename(tmp_dir / "01007EF00011E000")
+            if (tmp_dir / "dlc").exists():
+                (tmp_dir / "dlc").rename(tmp_dir / "01007EF00011F001")
         else:
             raise FileNotFoundError(
                 "This mod does not appear to have a valid folder structure"
@@ -329,11 +363,15 @@ def create_bnp_mod(mod: Path, output: Path, meta: dict, options: Optional[dict]
     meta["platform"] = (
         "any" if any_platform else "wiiu" if util.get_settings("wiiu") else "switch"
     )
+    if options["options"].get("general", {}).get("base_priority", False):
+        meta["priority"] = "base"
     (tmp_dir / "info.json").write_text(
         dumps(meta, ensure_ascii=False, indent=2), encoding="utf-8"
     )
 
-    with Pool() as pool:
+    _package_code(tmp_dir, meta)
+
+    with Pool(maxtasksperchild=500) as pool:
         yml_files = set(tmp_dir.glob("**/*.yml"))
         if yml_files:
             print("Compiling YAML documents...")
@@ -376,16 +414,17 @@ def create_bnp_mod(mod: Path, output: Path, meta: dict, options: Optional[dict]
                 f"There was an error generating change logs for your mod. {str(err)}"
             )
 
-        _clean_sarcs(tmp_dir, hashes, pool)
-        for folder in {d for d in tmp_dir.glob("options/*") if d.is_dir()}:
-            _clean_sarcs(folder, hashes, pool)
+        if not util.get_settings("strip_gfx"):
+            _clean_sarcs(tmp_dir, hashes, pool)
+            for folder in {d for d in tmp_dir.glob("options/*") if d.is_dir()}:
+                _clean_sarcs(folder, hashes, pool)
 
     print("Cleaning any junk files...")
     for file in {f for f in tmp_dir.rglob("**/*") if f.is_file()}:
         if "logs" in file.parts:
             continue
         if (
-            file.suffix in {".yml", ".json", ".bak", ".tmp", ".old"}
+            file.suffix in {".yml", ".json", ".bak", ".tmp", ".old", ".bnp"}
             and file.stem != "info"
         ):
             file.unlink()
@@ -396,6 +435,8 @@ def create_bnp_mod(mod: Path, output: Path, meta: dict, options: Optional[dict]
             shutil.rmtree(folder)
 
     print(f"Saving output file to {str(output)}...")
+    if output.exists():
+        output.unlink()
     x_args = [util.get_7z_path(), "a", str(output), f'{str(tmp_dir / "*")}']
     if system() == "Windows":
         subprocess.run(
@@ -418,7 +459,6 @@ NO_CONVERT_EXTS = {
     ".bfres",
     ".hkcl",
     ".hkrg",
-    ".bphyssb",
     ".sesetlist",
     ".sbfarc",
     ".shknm2",
@@ -434,7 +474,6 @@ NO_CONVERT_EXTS = {
     ".bfsar",
     ".sbmapopen",
     ".sbmaptex",
-    ".sbreviewtex",
 }
 
 
@@ -535,9 +574,12 @@ def _convert_sarc(sarc: oead.Sarc, to_wiiu: bool) -> Tuple[bytes, List[str]]:
 def convert_mod(mod: Path, to_wiiu: bool, warn_only: bool = False) -> list:
     warnings = []
 
-    def handle_warning(warning: str) -> None:
+    def handle_warning(warning: Union[str, list]) -> None:
         if not warn_only:
             raise ValueError(warning)
+        elif isinstance(warning, list):
+            for warning_ in warning:
+                warnings.append(warning_)
         else:
             warnings.append(warning)
 
@@ -558,25 +600,129 @@ def convert_mod(mod: Path, to_wiiu: bool, warn_only: bool = False) -> list:
 
     special_files = {"ActorInfo.product.sbyml"}
     all_files = {
-        f for f in mod.rglob("**/*.*") if f.is_file() and "options" not in f.parts
+        f for f in mod.rglob("*.*") if f.is_file() and f not in (mod / "options").rglob("*.*")
     }
 
     for file in all_files:
         if file.suffix in NO_CONVERT_EXTS:
             handle_warning(
                 "This mod contains a file which the platform converter does not support:"
-                f" {str(file.relative_to(mod))}"
+                f" {str(file.relative_to(mod)) if 'options' not in mod.parts else str(file.relative_to(mod.parent))}"
             )
 
     actorinfo_log = mod / "logs" / "actorinfo.yml"
     if actorinfo_log.exists():
         actorinfo = oead.byml.from_text(actorinfo_log.read_text("utf-8"))
-        if any("instSize" in actor for _, actor in actorinfo.items()):
-            handle_warning(
-                "This mod contains changes to actor instSize values, "
-                "which cannot be automatically converted."
+        from bcml.mergers.actors import get_stock_actorinfo
+
+        profiles = {
+            str(crc32(actor["name"].encode("utf-8"))): actor["profile"]
+            for actor in get_stock_actorinfo()["Actors"]
+        }
+        profile_ratios = {
+            "ComplexTag": 1.58580,
+            "ArmorLower": 2.77174,
+            "MapConstActive": 2.62619,
+            "Horse": 1.57119,
+            "MapConstPassive": 3.34795,
+            "Prey": 1.90845,
+            "MapDynamicActive": 2.54314,
+            "Anchor": 1.54027,
+            "NPC": 1.75036,
+            "DemoNPC": 2.44085,
+            "MapDynamicPassive": 2.87003,
+            "OptionalWeapon": 3.32726,
+            "ArmorUpper": 2.11312,
+            "ArmorHead": 2.27533,
+            "Item": 2.59118,
+            "WeaponBow": 2.43674,
+            "Enemy": 1.70107,
+            "CookResult": 1.47792,
+            "HorseSaddle": 2.41971,
+            "System": 1.59686,
+            "WeaponShield": 2.58867,
+            "MergedDungeonParts": 2.26925,
+            "NoCalcActor": 1.56243,
+            "SoundProxy": 1.57575,
+            "WeaponSmallSword": 2.56363,
+            "SiteBoss": 1.61253,
+            "ActorReaction": 1.95015,
+            "WeaponLargeSword": 2.40451,
+            "EffectLocater": 1.68679,
+            "Beam": 1.58558,
+            "SoleTag": 1.60180,
+            "GelEnemy": 1.69351,
+            "PlayerItem": 1.82444,
+            "WeaponSpear": 2.41490,
+            "Bullet": 2.12249,
+            "LinkTag": 1.67059,
+            "ActorOption": 1.47986,
+            "GiantEnemy": 1.64212,
+            "EventSystem": 1.58667,
+            "Dummy": 1.57355,
+            "Rope": 2.45780,
+            "Guardian": 2.00911,
+            "Swarm": 2.29493,
+            "GiantArmor": 2.74482,
+            "Dragon": 1.62425,
+            "GuardianComponent": 2.75823,
+            "Remains": 2.23782,
+            "CapturedActor": 2.33165,
+            "ArmorExtra1": 3.64020,
+            "AreaManagementForOthers": 1.56662,
+            "SpotBgmTag": 1.54055,
+            "Area": 1.57014,
+            "AreaManagementForPlayer": 1.57471,
+            "LineBeam": 1.58024,
+            "EnvSeEmitPoint": 1.59169,
+            "DemoEquipment": 2.78588,
+            "SoundSystemActor": 1.66232,
+            "HorseObject": 1.96967,
+            "Motorcycle": 1.89353,
+            "Sandworm": 1.66948,
+            "GameBlanceLocater": 2.71996,
+            "HorseReins": 2.78040,
+            "AirWall": 1.55907,
+            "NoWork": 1.55033,
+            "Weapon": 2.47555,
+            "SweepCollision": 1.53318,
+            "ArmorExtra0": 3.24801,
+            "Player": 1.44917,
+            "EnemySwarm": 2.03859,
+            "ArmorExtra2": 2.81442,
+            "WolfLink": 1.87220,
+            "PauseMenuPlayer": 1.86665,
+            "LastBoss": 1.68402,
+            "StaticAnchor": 1.53173,
+            "EditCamera": 1.59361,
+            "EventTag": 1.59889,
+            "GlobalParameter": 1.55041,
+            "Camera": 1.53002,
+        }
+        for hash_id, actor in [
+            (hash_id, actor)
+            for hash_id, actor in actorinfo.items()
+            if "instSize" in actor
+        ]:
+            profile = dict(actor).get("profile", profiles.get(hash_id))
+            if not profile:
+                handle_warning(
+                    f"Could not detect profile for actor with hash {hash_id}. "
+                    "The instSize value was not converted."
+                )
+                continue
+            ratio = profile_ratios[profile]
+            actor["instSize"] = oead.S32(
+                int(
+                    (
+                        (actor["instSize"].v * ratio)
+                        if not to_wiiu
+                        else (actor["instSize"].v / ratio)
+                    )
+                    * 1.1 # safety buffer, since we're dealing with averages
+                )
             )
-        del actorinfo
+        actorinfo_log.write_text(oead.byml.to_text(actorinfo))
 
     for log in {"drops.json", "packs.json"}:
         log_path = mod / "logs" / log
@@ -587,7 +733,7 @@ def convert_mod(mod: Path, to_wiiu: bool, warn_only: bool = False) -> list:
                 "utf-8",
             )
 
-    for log in {"deepmerge.aamp", "shop.aamp"}:
+    for log in {"deepmerge.aamp", "shop.aamp", "aslist.aamp"}:
         log_path = mod / "logs" / log
         if log_path.exists():
             pio = oead.aamp.ParameterIO.from_binary(log_path.read_bytes())
@@ -606,9 +752,12 @@ def convert_mod(mod: Path, to_wiiu: bool, warn_only: bool = False) -> list:
         f for f in all_files if f.suffix in BYML_EXTS and f.name not in special_files
     }:
         byml = oead.byml.from_binary(util.unyaz_if_needed(file.read_bytes()))
-        file.write_bytes(oead.byml.to_binary(byml, big_endian=to_wiiu))
+        data = oead.byml.to_binary(byml, big_endian=to_wiiu)
+        if file.suffix.startswith(".s"):
+            data = util.compress(data)
+        file.write_bytes(data)
 
-    with Pool() as pool:
+    with Pool(maxtasksperchild=500) as pool:
         errs = pool.map(
             partial(_convert_actorpack, to_wiiu=to_wiiu),
             {f for f in all_files if f.suffix == ".sbactorpack"},
@@ -651,9 +800,12 @@ def convert_mod(mod: Path, to_wiiu: bool, warn_only: bool = False) -> list:
 
     if (mod / "options").exists():
         for folder in {d for d in (mod / "options").glob("*") if d.is_dir()}:
-            convert_mod(folder, to_wiiu=to_wiiu, warn_only=warn_only)
+            option_warnings = convert_mod(folder, to_wiiu=to_wiiu, warn_only=warn_only)
+            handle_warning(option_warnings)
+    
+    if (mod / "info.json").exists():
+        meta = loads((mod / "info.json").read_text("utf-8"))
+        meta["platform"] = "wiiu" if to_wiiu else "switch"
+        (mod / "info.json").write_text(dumps(meta, indent=2, ensure_ascii=False), "utf-8")
 
-    meta = loads((mod / "info.json").read_text("utf-8"))
-    meta["platform"] = "wiiu" if to_wiiu else "switch"
-    (mod / "info.json").write_text(dumps(meta, indent=2, ensure_ascii=False), "utf-8")
     return warnings
