@@ -3,11 +3,13 @@
 # Licensed under GPLv3+
 # pylint: disable=too-many-lines
 import datetime
+import errno
 import json
 import multiprocessing
 import os
 import re
 import shutil
+import stat
 import subprocess
 from functools import partial
 from multiprocessing import Pool
@@ -25,14 +27,40 @@ from bcml.util import BcmlMod, get_7z_path
 
 
 def extract_mod_meta(mod: Path) -> Dict[str, Any]:
-    process = subprocess.Popen(
-        f'"{get_7z_path()}" e "{str(mod.resolve())}" -r -so info.json',
-        stdout=subprocess.PIPE,
-        shell=True,
-    )
-    out, _ = process.communicate()
-    process.wait()
-    return json.loads(out.decode("utf-8")) if out else {}
+    result: subprocess.CompletedProcess
+    if util.SYSTEM == "Windows":
+        result = subprocess.run(
+            [
+                get_7z_path(),
+                "e",
+                str(mod.resolve()),
+                "-r",
+                "-so",
+                "info.json",
+            ],
+            capture_output=True,
+            universal_newlines=True,
+            creationflags=util.CREATE_NO_WINDOW,
+        )
+    else:
+        result = subprocess.run(
+            [
+                get_7z_path(),
+                "e",
+                str(mod.resolve()),
+                "-r",
+                "-so",
+                "info.json",
+            ],
+            capture_output=True,
+            universal_newlines=True,
+        )
+    try:
+        assert not result.stderr
+        meta = json.loads(result.stdout)
+    except:
+        return {}
+    return meta
 
 
 def open_mod(path: Path) -> Path:
@@ -65,7 +93,11 @@ def open_mod(path: Path) -> Path:
             "or meta file (rules.txt or info.json)."
         )
     if not tmpdir.exists():
-        raise Exception("No files were extracted.")
+        raise Exception(
+            "No files were extracted. This may be because of an invalid or corrupted "
+            "download. If using the GameBanana tab, you may need to try again later, as"
+            " the problem could be caused by server errors."
+        )
 
     rulesdir = tmpdir
     if (rulesdir / "info.json").exists():
@@ -136,19 +168,26 @@ def find_modded_files(
         / "AocMainField.pack"
     )
     if aoc_field.exists() and aoc_field.stat().st_size > 0:
-        if not (
-            tmp_dir
-            / util.get_dlc_path()
-            / ("0010" if util.get_settings("wiiu") else "")
-        ).rglob("Map/**/?-?_*.smubin"):
+        if not list(
+            (
+                tmp_dir
+                / util.get_dlc_path()
+                / ("0010" if util.get_settings("wiiu") else "")
+            ).rglob("Map/**/?-?_*.smubin")
+        ):
             aoc_pack = oead.Sarc(aoc_field.read_bytes())
             for file in aoc_pack.get_files():
-                ex_out = tmp_dir / util.get_dlc_path() / "0010" / file.name
+                ex_out = (
+                    tmp_dir
+                    / util.get_dlc_path()
+                    / ("0010" if util.get_settings("wiiu") else "")
+                    / file.name
+                )
                 ex_out.parent.mkdir(parents=True, exist_ok=True)
                 ex_out.write_bytes(file.data)
         aoc_field.write_bytes(b"")
 
-    this_pool = pool or Pool()
+    this_pool = pool or Pool(maxtasksperchild=500)
     results = this_pool.map(
         partial(_check_modded, tmp_dir=tmp_dir),
         {
@@ -230,10 +269,12 @@ def generate_logs(
         options["disable"] = []
     util.vprint(options)
 
-    this_pool = pool or Pool()
+    this_pool = pool or Pool(maxtasksperchild=500)
     print("Scanning for modified files...")
     modded_files = find_modded_files(tmp_dir, pool=pool)
-    if not modded_files:
+    if not (
+        modded_files or (tmp_dir / "patches").exists() or (tmp_dir / "logs").exists()
+    ):
         raise RuntimeError(
             f"No modified files were found in {str(tmp_dir)}."
             "This probably means this mod is not in a supported format."
@@ -254,6 +295,10 @@ def generate_logs(
                 merger.set_options(options["options"][merger.NAME])
             merger.set_pool(this_pool)
             merger.log_diff(tmp_dir, modded_files)
+        if util.get_settings("strip_gfx"):
+            dev._clean_sarcs(
+                tmp_dir, util.get_hash_table(util.get_settings("wiiu")), this_pool
+            )
     except:  # pylint: disable=bare-except
         this_pool.close()
         this_pool.join()
@@ -288,6 +333,7 @@ def install_mod(
     pool: Optional[multiprocessing.pool.Pool] = None,
     insert_priority: int = 0,
     merge_now: bool = False,
+    updated: bool = False,
 ):
     if not insert_priority:
         insert_priority = get_next_priority()
@@ -325,11 +371,14 @@ def install_mod(
         mod_name = rules["name"].strip(" '\"").replace("_", "")
         print(f"Identified mod: {mod_name}")
         if rules["depends"]:
-            installed_metas = {
-                v[0]: v[1]
-                for m in util.get_installed_mods()
-                for v in util.BcmlMod.meta_from_id(m.id)
-            }
+            try:
+                installed_metas = {
+                    v[0]: v[1]
+                    for m in util.get_installed_mods()
+                    for v in util.BcmlMod.meta_from_id(m.id)
+                }
+            except (IndexError, TypeError) as err:
+                raise RuntimeError(f"This BNP has invalid or corrupt dependency data.")
             for depend in rules["depends"]:
                 depend_name, depend_version = util.BcmlMod.meta_from_id(depend)
                 if (depend_name not in installed_metas) or (
@@ -348,6 +397,8 @@ def install_mod(
                 f'"{mod_name}" is for {friendly_plaform(rules["platform"])}, not '
                 f" {friendly_plaform(user_platform)}.'"
             )
+        if "priority" in rules and rules["priority"] == "base":
+            insert_priority = 100
 
         logs = tmp_dir / "logs"
         if logs.exists():
@@ -360,9 +411,13 @@ def install_mod(
                 if merger.is_mod_logged(BcmlMod(tmp_dir)):
                     (tmp_dir / "logs" / merger.log_name).unlink()
         else:
-            this_pool = pool or Pool()
-            generate_logs(tmp_dir=tmp_dir, options=options, pool=pool)
-            (tmp_dir / ".processed").touch()
+            this_pool = pool or Pool(maxtasksperchild=500)
+            dev._pack_sarcs(
+                tmp_dir, util.get_hash_table(util.get_settings("wiiu")), this_pool
+            )
+            generate_logs(tmp_dir=tmp_dir, options=options, pool=this_pool)
+            if not util.get_settings("strip_gfx"):
+                (tmp_dir / ".processed").touch()
     except Exception as err:  # pylint: disable=broad-except
         try:
             name = mod_name
@@ -370,7 +425,7 @@ def install_mod(
             name = "your mod, the name of which could not be detected"
         raise util.InstallError(err, name) from err
 
-    if selects:
+    if selects is not None:
         for opt_dir in {d for d in (tmp_dir / "options").glob("*") if d.is_dir()}:
             if opt_dir.name not in selects:
                 shutil.rmtree(opt_dir, ignore_errors=True)
@@ -432,9 +487,23 @@ def install_mod(
     mod_dir = util.get_modpack_dir() / mod_id
 
     try:
-        for existing_mod in util.get_installed_mods(True):
-            if existing_mod.priority >= priority:
-                existing_mod.change_priority(existing_mod.priority + 1)
+        if not updated:
+            for existing_mod in util.get_installed_mods(True):
+                if existing_mod.priority >= priority:
+                    existing_mod.change_priority(existing_mod.priority + 1)
+
+        if (tmp_dir / "patches").exists() and not util.get_settings("no_cemu"):
+            patch_dir = (
+                util.get_cemu_dir()
+                / "graphicPacks"
+                / f"bcmlPatches"
+                / util.get_safe_pathname(rules["name"])
+            )
+            patch_dir.mkdir(parents=True, exist_ok=True)
+            for file in {f for f in (tmp_dir / "patches").rglob("*") if f.is_file()}:
+                out = patch_dir / file.relative_to(tmp_dir / "patches")
+                out.parent.mkdir(parents=True, exist_ok=True)
+                shutil.copyfile(file, out)
 
         mod_dir.parent.mkdir(parents=True, exist_ok=True)
         print(f"Moving mod to {str(mod_dir)}...")
@@ -453,6 +522,7 @@ def install_mod(
                     )
         elif mod.is_dir():
             shutil.copytree(str(tmp_dir), str(mod_dir))
+
         shutil.rmtree(tmp_dir, ignore_errors=True)
 
         rules["priority"] = priority
@@ -469,8 +539,6 @@ def install_mod(
             util.get_mod_preview(output_mod)
         except Exception:  # pylint: disable=broad-except
             pass
-
-        print(f"Enabling {mod_name} in Cemu...")
     except Exception as err:  # pylint: disable=broad-except
         if mod_dir.exists():
             try:
@@ -481,14 +549,12 @@ def install_mod(
 
     try:
         if merge_now:
-            all_mergers = set()
-            for merger in {m() for m in mergers.get_mergers()}:  # type: ignore
-                if merger.is_mod_logged(output_mod):
-                    all_mergers.add(merger)
-            if this_pool:
-                for merger in mergers.sort_mergers(all_mergers):
-                    merger.set_pool(this_pool)
-                    merger.perform_merge()
+            for merger in [m() for m in mergers.get_mergers()]:
+                if this_pool or pool:
+                    merger.set_pool(this_pool or pool)
+                if merger.NAME in options["options"]:
+                    merger.set_options(options["options"][merger.NAME])
+                merger.perform_merge()
     except Exception as err:  # pylint: disable=broad-except
         raise util.MergeError(err) from err
 
@@ -522,12 +588,27 @@ def enable_mod(mod: BcmlMod, wait_merge: bool = False):
     print(f"{mod.name} enabled")
 
 
+def force_del(func, path, exc):
+    excvalue = exc[1]
+    if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
+        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
+        func(path)
+    else:
+        raise RuntimeError(
+            f"The folder {path} could not be removed. "
+            "You may need to delete it manually and remerge, or "
+            "close all open programs (including BCML and Windows Explorer) "
+            "and try again. The location of the folder is "
+            f"<code>{str(path)}</code>."
+        )
+
+
 @refresher
 def uninstall_mod(mod: BcmlMod, wait_merge: bool = False):
-    print(f"Uninstalling {mod.name}...")
+    has_patches = (mod.path / "patches").exists()
     try:
-        shutil.rmtree(str(mod.path))
-    except (OSError, PermissionError) as err:
+        shutil.rmtree(str(mod.path), onerror=force_del)
+    except (OSError, PermissionError, WindowsError) as err:
         raise RuntimeError(
             f"The folder for {mod.name} could not be removed. "
             "You may need to delete it manually and remerge, or "
@@ -548,6 +629,15 @@ def uninstall_mod(mod: BcmlMod, wait_merge: bool = False):
         if not wait_merge:
             refresh_merges()
 
+    if has_patches and not util.get_settings("no_cemu"):
+        shutil.rmtree(
+            util.get_cemu_dir()
+            / "graphicPacks"
+            / "bcmlPatches"
+            / util.get_safe_pathname(mod.name),
+            ignore_errors=True,
+        )
+
     print(f"{mod.name} has been uninstalled.")
 
 
@@ -555,7 +645,7 @@ def refresh_merges():
     print("Cleansing old merges...")
     shutil.rmtree(util.get_master_modpack_dir(), True)
     print("Refreshing merged mods...")
-    with Pool() as pool:
+    with Pool(maxtasksperchild=500) as pool:
         for merger in mergers.sort_mergers(
             [merger_class() for merger_class in mergers.get_mergers()]
         ):
@@ -620,50 +710,53 @@ def restore_backup(backup: Union[str, Path]):
 
 
 def enable_bcml_gfx():
-    if not util.get_settings("no_cemu"):
-        settings = util.parse_cemu_settings()
-        try:
-            gpack = settings.getElementsByTagName("GraphicPack")[0]
-        except IndexError:
-            gpack = settings.createElement("GraphicPack")
-            settings.appendChild(gpack)
-        new_cemu = True
-        entry: minidom.Element
-        for entry in gpack.getElementsByTagName("Entry"):
-            if new_cemu and entry.getElementsByTagName("filename"):
-                new_cemu = False
+    if util.get_settings("no_cemu"):
+        return
+
+    settings = util.parse_cemu_settings()
+    try:
+        gpack = settings.getElementsByTagName("GraphicPack")[0]
+    except IndexError:
+        gpack = settings.createElement("GraphicPack")
+        settings.appendChild(gpack)
+    new_cemu = True
+
+    def create_entry(path: str):
+        def entry_matches(entry):
             try:
-                if (
-                    "BCML"
-                    in entry.getElementsByTagName("filename")[0].childNodes[0].data
-                ):
-                    break
+                return (
+                    path == entry.getElementsByTagName("filename")[0].childNodes[0].data
+                )
             except IndexError:
-                if "BCML" in entry.getAttribute("filename"):
-                    break
+                return path == entry.getAttribute("filename")
+
+        if any(entry_matches(entry) for entry in gpack.getElementsByTagName("Entry")):
+            return
+        entry: minidom.Element = settings.createElement("Entry")
+        if new_cemu:
+            entry.setAttribute("filename", path)
         else:
-            bcmlentry = settings.createElement("Entry")
-            if new_cemu:
-                bcmlentry.setAttribute(
-                    "filename", "graphicPacks\\BreathOfTheWild_BCML\\rules.txt"
-                )
-            else:
-                entryfile = settings.createElement("filename")
-                entryfile.appendChild(
-                    settings.createTextNode(
-                        "graphicPacks\\BreathOfTheWild_BCML\\rules.txt"
-                    )
-                )
-                bcmlentry.appendChild(entryfile)
-            entrypreset = settings.createElement("preset")
-            entrypreset.appendChild(settings.createTextNode(""))
-            bcmlentry.appendChild(entrypreset)
-            gpack.appendChild(bcmlentry)
-            settings.writexml(
-                (util.get_cemu_dir() / "settings.xml").open("w", encoding="utf-8"),
-                addindent="    ",
-                newl="\n",
-            )
+            entryfile = settings.createElement("filename")
+            entryfile.appendChild(settings.createTextNode(path))
+            entry.appendChild(entryfile)
+        entrypreset = settings.createElement("preset")
+        entrypreset.appendChild(settings.createTextNode(""))
+        entry.appendChild(entrypreset)
+        gpack.appendChild(entry)
+
+    create_entry("graphicPacks\\BreathOfTheWild_BCML\\rules.txt")
+
+    if (util.get_cemu_dir() / "graphicPacks" / "bcmlPatches").exists():
+        for rules in (util.get_cemu_dir() / "graphicPacks" / "bcmlPatches").rglob(
+            "rules.txt"
+        ):
+            create_entry(str(rules.relative_to(util.get_cemu_dir())))
+
+        settings.writexml(
+            (util.get_cemu_dir() / "settings.xml").open("w", encoding="utf-8"),
+            addindent="    ",
+            newl="\n",
+        )
 
 
 def disable_bcml_gfx():
@@ -681,15 +774,15 @@ def disable_bcml_gfx():
                 new_cemu = False
             try:
                 if (
-                    "BCML"
-                    in entry.getElementsByTagName("filename")[0].childNodes[0].data
+                    "bcml"
+                    in entry.getElementsByTagName("filename")[0]
+                    .childNodes[0]
+                    .data.lower()
                 ):
                     gpack.removeChild(entry)
-                    break
             except IndexError:
-                if "BCML" in entry.getAttribute("filename"):
+                if "bcml" in entry.getAttribute("filename").lower():
                     gpack.removeChild(entry)
-                    break
         settings.writexml(
             (util.get_cemu_dir() / "settings.xml").open("w", encoding="utf-8"),
             addindent="    ",
@@ -700,14 +793,32 @@ def disable_bcml_gfx():
 def link_master_mod(output: Path = None):
     util.create_bcml_graphicpack_if_needed()
     if not output:
-        if util.get_settings("no_cemu"):
+        if not util.get_settings(
+            "export_dir" if util.get_settings("wiiu") else "export_dir_nx"
+        ):
             return
-        output = util.get_cemu_dir() / "graphicPacks" / "BreathOfTheWild_BCML"
+        output = Path(
+            util.get_settings(
+                "export_dir" if util.get_settings("wiiu") else "export_dir_nx"
+            )
+        )
     if output.exists():
-        shutil.rmtree(output, ignore_errors=True)
+        if not util.get_settings("no_cemu"):
+            shutil.rmtree(output, ignore_errors=True)
+        elif util.get_settings("wiiu"):
+            if (output / "content").exists():
+                shutil.rmtree((output / "content"), ignore_errors=True)
+            if (output / "aoc").exists():
+                shutil.rmtree((output / "aoc"), ignore_errors=True)
+        elif not util.get_settings("wiiu"):
+            if (output / "01007EF00011E000").exists():
+                shutil.rmtree((output / "01007EF00011E000"), ignore_errors=True)
+            if (output / "01007EF00011F001").exists():
+                shutil.rmtree((output / "01007EF00011F001"), ignore_errors=True)
     try:
         output.mkdir(parents=True, exist_ok=True)
         if not util.get_settings("no_cemu"):
+            output.mkdir(parents=True, exist_ok=True)
             shutil.copy(
                 util.get_master_modpack_dir() / "rules.txt", output / "rules.txt"
             )
@@ -715,8 +826,11 @@ def link_master_mod(output: Path = None):
         raise RuntimeError(
             "There was a problem creating the master BCML graphic pack. "
             "It may be a one time fluke, so try remerging and/or restarting BCML. "
-            "If the problem persists, good luck, because it's something wonky about your "
-            "PC, I guess."
+            "This can also happen if BCML and/or Cemu are installed into Program "
+            "Files or any folder which requires administrator (or root) permissions. "
+            "You can try running BCML as administrator or root, but bear in mind this "
+            "is not officially supported. If the problem persists, good luck, because "
+            "it's something wonky about your PC, I guess."
         ) from err
 
     mod_folders: List[Path] = sorted(
@@ -754,9 +868,20 @@ def link_master_mod(output: Path = None):
                         link_or_copy(str(item), str(output / rel_path))
                     else:
                         raise
+    if len(list(output.rglob("*"))) == 0:
+        raise RuntimeError(
+            "No files were created in your export directory. This may mean BCML"
+            " lacked then necessary permissions to write to the folder, or "
+            "the folder is otherwise unusable by Python. (This often happens"
+            " if the target folder is connected to OneDrive, WinFUSE, or "
+            "perhaps network drives._ If possible, try changing your export "
+            "folder or its permissions. If you are a Cemu user and your Cemu "
+            "folder is in your OneDrive, you will probably need to move Cemu "
+            "somewhere else."
+        )
 
 
-def export(output: Path):
+def export(output: Path, standalone: bool = False):
     print("Loading files...")
     tmp_dir = Path(mkdtemp())
     if tmp_dir.exists():
@@ -772,12 +897,18 @@ def export(output: Path):
     rules_path = tmp_dir / "rules.txt"
     mods = util.get_installed_mods()
     if util.get_settings("wiiu"):
+        name = "Exported BCML mod" if not standalone else mods[0].name
+        desc = (
+            f'Exported merge of {", ".join([mod.name for mod in mods])}'
+            if not standalone
+            else mods[0].description
+        )
         rules_path.write_text(
             "[Definition]\n"
             "titleIds = 00050000101C9300,00050000101C9400,00050000101C9500\n"
-            "name = Exported BCML Mod\n"
-            "path = The Legend of Zelda: Breath of the Wild/Mods/Exported BCML\n"
-            f'description = Exported merge of {", ".join([mod.name for mod in mods])}\n'
+            f"name = {name}\n"
+            f"path = The Legend of Zelda: Breath of the Wild/Mods/{name}\n"
+            f"description = {desc}\n"
             "version = 4\n",
             encoding="utf-8",
         )
