@@ -84,12 +84,13 @@ def diff_gamedata_type(data_type: str, mod_data: dict, stock_data: dict) -> Hash
 
 
 def get_modded_gamedata_entries(gamedata: oead.Sarc, pool: pool.Pool = None) -> Hash:
-    this_pool = pool or Pool()
+    this_pool = pool or Pool(maxtasksperchild=500)
     stock_data = consolidate_gamedata(get_stock_gamedata())
     mod_data = consolidate_gamedata(gamedata)
     del gamedata
     results = this_pool.starmap(
-        diff_gamedata_type, ((key, mod_data[key], stock_data[key]) for key in mod_data),
+        diff_gamedata_type,
+        ((key, mod_data[key], stock_data[key]) for key in mod_data),
     )
     diffs = Hash({data_type: diff for d in results for data_type, diff in d.items()})
     del results
@@ -110,18 +111,22 @@ def get_modded_savedata_entries(savedata: oead.Sarc) -> Hash:
     }
     new_entries = oead.byml.Array()
     mod_hashes = set()
-    for file in sorted(savedata.get_files(), key=lambda f: f.name,)[0:-2]:
-        entries = oead.byml.from_binary(file.data)["file_list"][1]
+    for file in savedata.get_files():
+        data = oead.byml.from_binary(file.data)
+        if data["file_list"][0]["file_name"] != "game_data.sav":
+            continue
+        entries = data["file_list"][1]
         mod_hashes |= {int(item["HashValue"]) for item in entries}
         new_entries.extend(
-            {item for item in entries if int(item["HashValue"]) not in ref_hashes}
+            [item for item in entries if int(item["HashValue"]) not in ref_hashes]
         )
     del ref_savedata
     return Hash(
         {
             "add": new_entries,
             "del": oead.byml.Array(
-                {oead.S32(item) for item in ref_hashes if item not in mod_hashes}
+                oead.S32(item)
+                for item in {item for item in ref_hashes if item not in mod_hashes}
             ),
         }
     )
@@ -256,7 +261,7 @@ class GameDataMerger(mergers.Merger):
 
         merged_entries = Hash(
             {
-                data_type: oead.byml.Array({value for _, value in entries.items()})
+                data_type: oead.byml.Array([value for _, value in entries.items()])
                 for data_type, entries in merged_entries.items()
             }
         )
@@ -390,9 +395,9 @@ class SaveDataMerger(mergers.Merger):
         hashes = set()
         for diff in reversed(diffs):
             for entry in diff["add"]:
-                if entry["HashValue"] not in hashes:
+                if entry["HashValue"].v not in hashes:
                     all_diffs["add"].append(entry)
-                hashes.add(entry["HashValue"])
+                hashes.add(entry["HashValue"].v)
             for entry in diff["del"]:
                 if entry not in all_diffs["del"]:
                     all_diffs["del"].append(entry)
@@ -420,6 +425,7 @@ class SaveDataMerger(mergers.Merger):
 
         savedata = get_stock_savedata()
         save_files = sorted(savedata.get_files(), key=lambda f: f.name)[0:-2]
+        del_ids = {item.v for item in new_entries["del"]}
 
         print("Merging changes...")
         merged_entries = oead.byml.Array(
@@ -434,7 +440,7 @@ class SaveDataMerger(mergers.Merger):
                         ],
                         *new_entries["add"],
                     ]
-                    if entry not in new_entries["del"]
+                    if entry["HashValue"].v not in del_ids
                 }.values(),
                 key=itemgetter("HashValue"),
             )
