@@ -1,6 +1,8 @@
 # pylint: disable=missing-docstring,no-member,too-many-lines,invalid-name
 # Copyright 2020 Nicene Nerd <macadamiadaze@gmail.com>
 # Licensed under GPLv3+
+import functools
+import gc
 import json
 import os
 import re
@@ -22,7 +24,7 @@ from multiprocessing import current_process
 from pathlib import Path
 from platform import system, python_version_tuple
 from pprint import pformat
-from subprocess import run
+from subprocess import run, PIPE
 from tempfile import mkdtemp
 from time import time_ns
 from typing import Union, List, Dict, ByteString, Tuple, Any, Optional, IO
@@ -32,9 +34,10 @@ import oead
 import requests
 import xxhash  # pylint: disable=wrong-import-order
 from oead.aamp import ParameterIO, ParameterList  # pylint:disable=import-error
-from webviewb import Window  # pylint: disable=wrong-import-order
+from webview import Window  # pylint: disable=wrong-import-order
 
 from bcml import pickles, DEBUG  # pylint: disable=unused-import
+from bcml.__version__ import VERSION
 
 
 CREATE_NO_WINDOW = 0x08000000
@@ -44,13 +47,13 @@ SARC_EXTS = {
     ".bactorpack",
     ".bmodelsh",
     ".beventpack",
-    ".stera",
+    # ".stera",
     ".stats",
     ".ssarc",
     ".sbactorpack",
     ".sbmodelsh",
     ".sbeventpack",
-    ".sstera",
+    # ".sstera",
     ".sstats",
     ".sblarc",
     ".blarc",
@@ -265,9 +268,7 @@ class BcmlMod:
     def __init__(self, mod_path):
         self.path = mod_path
         try:
-            self._info = json.loads(
-                (self.path / "info.json").read_text("utf-8"), encoding="utf-8"
-            )
+            self._info = json.loads((self.path / "info.json").read_text("utf-8"))
             assert "name" in self._info
             assert "id" in self._info
             assert "priority" in self._info
@@ -465,6 +466,14 @@ def timed(func):
     return timed_function
 
 
+def clear_all_caches():
+    gc.collect()
+    for wrap in {
+        a for a in gc.get_objects() if isinstance(a, functools._lru_cache_wrapper)
+    }:
+        wrap.cache_clear()
+
+
 def sanity_check():
     ver = python_version_tuple()
     if int(ver[0]) < 3 or (int(ver[0]) >= 3 and int(ver[1]) < 7):
@@ -490,11 +499,24 @@ def get_exec_dir() -> Path:
     return Path(os.path.dirname(os.path.realpath(__file__)))
 
 
-@lru_cache(1)
-def get_python_exe() -> str:
-    if get_exec_dir().parent.name == "pkgs":
-        return str(get_exec_dir().parent.parent / "Python" / "python.exe")
-    return sys.executable
+@lru_cache(2)
+def get_python_exe(gui: bool) -> Path:
+    embedded = [
+        d
+        for d in get_exec_dir().parents
+        if (d / "python37._pth").exists() and (d / "pythonw.exe").exists()
+    ]
+    if embedded:
+        return embedded[0] / ("pythonw.exe" if gui else "python.exe")
+    else:
+        if SYSTEM == "Windows":
+            return (
+                sys.executable.replace("pythonw.exe", "python.exe")
+                if not gui
+                else sys.executable.replace("python.exe", "pythonw.exe")
+            )
+        else:
+            return sys.executable
 
 
 @lru_cache(None)
@@ -546,6 +568,8 @@ DEFAULT_SETTINGS = {
     "dlc_dir": "",
     "dlc_dir_nx": "",
     "store_dir": str(get_data_dir()),
+    "export_dir": "",
+    "export_dir_nx": "",
     "load_reverse": False,
     "site_meta": "",
     "dark_theme": False,
@@ -556,6 +580,12 @@ DEFAULT_SETTINGS = {
     "no_hardlinks": False,
     "force_7z": False,
     "suppress_update": False,
+    "nsfw": False,
+    "last_version": VERSION,
+    "changelog": True,
+    "strip_gfx": False,
+    "auto_gb": True,
+    "show_gb": False,
 }
 
 
@@ -569,12 +599,22 @@ def get_settings(name: str = "") -> Any:
                 with settings_path.open("w", encoding="utf-8") as s_file:
                     json.dump(settings, s_file)
             else:
-                settings.update(json.loads(settings_path.read_text()))
+                tmp_settings_path = get_data_dir() / "tmp_settings.json"
+                if tmp_settings_path.exists():
+                    settings.update(json.loads(tmp_settings_path.read_text()))
+                else:
+                    settings.update(json.loads(settings_path.read_text()))
                 for k, v in DEFAULT_SETTINGS.items():
                     if k not in settings:
                         settings[k] = v
                 if settings["store_dir"] == "":
                     settings["store_dir"] = str(get_data_dir())
+                if settings["cemu_dir"] and not settings["no_cemu"]:
+                    settings["export_dir"] = str(
+                        Path(settings["cemu_dir"])
+                        / "graphicPacks"
+                        / "BreathOfTheWild_BCML"
+                    )
             setattr(get_settings, "settings", settings)
         if name:
             return getattr(get_settings, "settings", {}).get(name, False)
@@ -732,18 +772,11 @@ def get_dlc_path() -> str:
     return "aoc" if get_settings("wiiu") else "01007EF00011F001/romfs"
 
 
-class TempModContext(AbstractContextManager):
-    _tmpdir: Path
-
-    def __init__(self, path: Path = None):
-        self._tmpdir = path or Path(mkdtemp())
-
-    def __enter__(self):
-        setattr(get_modpack_dir, "tmp", self._tmpdir)
-
-    def __exit__(self, exctype, excinst, exctb):
-        delattr(get_modpack_dir, "tmp")
-        shutil.rmtree(self._tmpdir, ignore_errors=True)
+def get_user_languages(folder: Optional[Path] = None) -> set:
+    langs = set()
+    for file in ((folder or get_update_dir()) / "Pack").glob("Bootup_????.pack"):
+        langs.add(get_file_language(file.name))
+    return langs
 
 
 class TempSettingsContext(AbstractContextManager):
@@ -755,18 +788,37 @@ class TempSettingsContext(AbstractContextManager):
         self._tmp_settings = tmp_settings
 
     def __enter__(self):
+        clear_all_caches()
         getattr(get_settings, "settings").update(self._tmp_settings)
+        with (get_data_dir() / "tmp_settings.json").open(
+            "w", encoding="utf-8"
+        ) as s_file:
+            json.dump(get_settings.settings, s_file, indent=2)
 
     def __exit__(self, exctype, excinst, exctb):
         setattr(get_settings, "settings", self._settings)
+        clear_all_caches()
+        (get_data_dir() / "tmp_settings.json").unlink()
+
+
+class TempModContext(TempSettingsContext):
+    _tmpdir: Path
+
+    def __init__(self, path: Path = None):
+        self._tmpdir = path or Path(mkdtemp())
+        super().__init__({"store_dir": str(self._tmpdir)})
+
+    def __exit__(self, exctype, excinst, exctb):
+        shutil.rmtree(self._tmpdir, ignore_errors=True)
+        super().__exit__(exctype, excinst, exctb)
 
 
 def get_modpack_dir() -> Path:
-    return getattr(
-        get_modpack_dir,
-        "tmp",
-        get_storage_dir() / ("mods" if get_settings("wiiu") else "mods_nx"),
-    )
+    return get_storage_dir() / ("mods" if get_settings("wiiu") else "mods_nx")
+
+
+def get_profiles_dir() -> Path:
+    return get_storage_dir() / ("profiles" if get_settings("wiiu") else "profiles_nx")
 
 
 @lru_cache(None)
@@ -907,6 +959,9 @@ def get_file_language(file: Union[Path, str]) -> str:
 
 
 def is_file_modded(name: str, file: Union[bytes, Path], count_new: bool = True) -> bool:
+    table = get_hash_table(get_settings("wiiu"))
+    if name not in table:
+        return count_new
     contents = (
         file
         if isinstance(file, bytes)
@@ -919,9 +974,6 @@ def is_file_modded(name: str, file: Union[bytes, Path], count_new: bool = True)
             contents = decompress(contents)
         except RuntimeError as err:
             raise ValueError(f"Invalid yaz0 file {name}") from err
-    table = get_hash_table(get_settings("wiiu"))
-    if name not in table:
-        return count_new
     fhash = xxhash.xxh64_intdigest(contents)
     return not fhash in table[name]
 
@@ -983,7 +1035,14 @@ def inject_files_into_actor(actor: str, files: Dict[str, ByteString]):
             / f"{actor}.sbactorpack"
         )
         if not actor_path.exists():
-            actor_path = get_game_file(f"Actor/Pack/{actor}.sbactorpack")
+            for m in reversed(get_installed_mods()):
+                try:
+                    actor_path = list(m.path.rglob(f"{actor}.sbactorpack"))[0]
+                    break
+                except IndexError:
+                    continue
+            else:
+                actor_path = get_game_file(f"Actor/Pack/{actor}.sbactorpack")
         actor_sarc = oead.Sarc(decompress(actor_path.read_bytes()))
     new_sarc = oead.SarcWriter.from_sarc(actor_sarc)
     del actor_sarc
@@ -1142,12 +1201,64 @@ def create_bcml_graphicpack_if_needed():
             "name = BCML\n"
             "path = The Legend of Zelda: Breath of the Wild/Mods/BCML\n"
             "description = Complete pack of mods merged using BCML\n"
-            "version = 4\n"
+            "version = 7\n"
+            "default = true\n"
             "fsPriority = 9999",
             encoding="utf-8",
         )
 
 
+def create_shortcuts(desktop: bool, start_menu: bool):
+    from pycrosskit.shortcuts import Shortcut
+
+    Shortcut(
+        "BCML",
+        shutil.which("bcml"),
+        "BOTW Cross-Platform Mod Loader",
+        str(get_exec_dir() / "data" / "bcml.ico"),
+        desktop,
+        start_menu,
+    )
+
+
+def download_webview2():
+    from bcml import native_msg
+
+    native_msg(
+        "Neither CEF nor Edge WebView2 is available. "
+        " Click OK to download the WebView2 runtime at https://go.microsoft.com/fwlink/p/?LinkId=2124703. "
+        " The download may take a few moments. Once done, install it and restart BCML.",
+        "Error",
+    )
+    path = Path(mkdtemp()) / f"webview.exe"
+    try:
+        res: requests.Response = requests.get(
+            "https://go.microsoft.com/fwlink/p/?LinkId=2124703"
+        )
+        with path.open("wb") as tmp_file:
+            for chunk in res.iter_content(chunk_size=1024):
+                tmp_file.write(chunk)
+    except (
+        FileNotFoundError,
+        PermissionError,
+        OSError,
+        requests.ConnectionError,
+        requests.RequestException,
+    ) as err:
+        native_msg(str(err), "Error")
+    try:
+        run(
+            [str(path)],
+            stdout=PIPE,
+            stderr=PIPE,
+            creationflags=CREATE_NO_WINDOW,
+            check=True,
+        )
+    except Exception as err:
+        native_msg(str(err), "Error")
+    sys.exit(0)
+
+
 UNDERRIDE = "UNDERRIDE_CONST"
 
 
@@ -1165,7 +1276,7 @@ def dict_merge(
             and (isinstance(dct[k], (dict, oead.byml.Hash)))
             and (isinstance(merge_dct[k], (Mapping, oead.byml.Hash)))
         ):
-            dict_merge(dct[k], merge_dct[k])
+            dict_merge(dct[k], merge_dct[k], overwrite_lists=overwrite_lists)
         elif (
             k in dct
             and (isinstance(dct[k], (list, oead.byml.Array)))
@@ -1242,30 +1353,20 @@ def get_open_port():
 
 @lru_cache(1)
 def get_latest_bcml() -> str:
-    args = [
-        get_python_exe().replace("pythonw", "python"),
-        "-m",
-        "pip",
-        "--disable-pip-version-check",
-        "install",
-        "bcml==checkver",
-    ]
-    if SYSTEM == "Windows":
-        result = run(
-            args,
-            creationflags=CREATE_NO_WINDOW,
-            capture_output=True,
-            universal_newlines=True,
-            check=False,
-        )
-    else:
-        result = run(args, capture_output=True, universal_newlines=True, check=False)
-    vers = sorted(re.findall(r"[0-9]\.[0-9]+\.[0-9a-z]+", result.stderr))
-    if not DEBUG:
-        vers = [v for v in vers if ".a" not in v and ".b" not in v]
     try:
-        return vers[-1]
-    except IndexError:
+        res = requests.get("https://pypi.org/rss/project/bcml/releases.xml")
+        doc = minidom.parseString(res.text)
+        versions = sorted(
+            (
+                item.getElementsByTagName("title")[0].childNodes[0].data
+                for item in doc.getElementsByTagName("item")
+            ),
+            reverse=True,
+        )
+        if DEBUG:
+            return versions[0]
+        return next(v for v in versions if "a" not in v and "b" not in v)
+    except:  # pylint: disable=bare-except
         return "0.0.0"
 
 
@@ -1340,7 +1441,10 @@ def get_7z_path():
     if system() == "Windows":
         return str(get_exec_dir() / "helpers" / "7z.exe")
     bundle_path = get_exec_dir() / "helpers" / "7z"
-    os.chmod(bundle_path, 0o755)
+    if not os.access(bundle_path, os.X_OK):
+        if not os.access(bundle_path, os.W_OK):
+            raise PermissionError(f"{bundle_path} is not executable and we don't have the permissions to change that")
+        os.chmod(bundle_path, 0o755)
     if get_settings("force_7z"):
         return str(bundle_path)
     return shutil.which("7z") or str(bundle_path)
